@@ -21,10 +21,10 @@ A simple Solr client for python.
 
 Features
 --------
- * Supports SOLR 1.2+
+ * Supports Solr 1.2+
  * Supports http/https and SSL client-side certificates
  * Uses persistent HTTP connections by default
- * Properly converts to/from SOLR data types, including datetime objects
+ * Properly converts to/from Solr data types, including datetime objects
  * Supports both querying and update commands (add, delete)
  * Requires Python 2.3+
 
@@ -34,7 +34,7 @@ Connections
 `SolrConnection` can be passed in the following parameters.
 Only `url` is required,.
 
-    url -- URI pointing to the SOLR instance. Examples:
+    url -- URI pointing to the Solr instance. Examples:
 
         http://localhost:8080/solr
         https://solr-server/solr
@@ -52,16 +52,19 @@ Only `url` is required,.
     ssl_key, ssl_cert -- If using client-side key files for
         SSL authentication,  these should be, respectively,
         your PEM key file and certificate file
+        
+    http_user, http_pass -- If given, include HTTP Basic authentication 
+        in all request headers.
 
 Once created, a connection object has the following public methods:
 
-    query (q, fields=None, highlight=None,
-           score=True, sort=None, **params)
+    query(q, fields=None, highlight=None,
+          score=True, sort=None, **params)
 
             q -- the query string.
 
             fields -- optional list of fields to include. It can be either
-                a string in the format that SOLR expects ('id,f1,f2'), or
+                a string in the format that Solr expects ('id,f1,f2'), or
                 a python list/tuple of field names.   Defaults to returning
                 all fields. ("*")
 
@@ -77,10 +80,10 @@ Once created, a connection object has the following public methods:
 
             sort -- list of fields to sort by.
 
-            Any parameters available to SOLR 'select' calls can also be
+            Any parameters available to Solr 'select' calls can also be
             passed in as named parameters (e.g., fq='...', rows=20, etc).
 
-            Many SOLR parameters are in a dotted notation (e.g.,
+            Many Solr parameters are in a dotted notation (e.g.,
             `hl.simple.post`).  For such parameters, replace the dots with
             underscores when calling this method. (e.g.,
             hl_simple_post='</pre'>)
@@ -137,11 +140,11 @@ Once created, a connection object has the following public methods:
     raw_query(**params)
 
             Send a query command (unprocessed by this library) to
-            the SOLR server. The resulting text is returned un-parsed.
+            the Solr server. The resulting text is returned un-parsed.
 
                 raw_query(q='id:1', wt='python', indent='on')
 
-            Many SOLR parameters are in a dotted notation (e.g.,
+            Many Solr parameters are in a dotted notation (e.g.,
             `hl.simple.post`).  For such parameters, replace the dots with
             underscores when calling this method. (e.g.,
             hl_simple_post='</pre'>)
@@ -177,8 +180,8 @@ Query Responses
 
             header['status'] -- status code.
 
-            See SOLR documentation for other/typical return values.
-            This may be settable at the SOLR-level in your config files.
+            See Solr documentation for other/typical return values.
+            This may be settable at the Solr-level in your config files.
 
 
         next_batch() -- If only a partial set of matches were returned
@@ -250,9 +253,10 @@ from xml.sax.handler import ContentHandler
 from xml.sax.saxutils import escape, quoteattr
 from xml.dom.minidom import parseString
 
-__version__ = "0.9.1"
+__version__ = "0.9.2"
 
-__all__ = ['SolrException', 'SolrConnection', 'Response']
+__all__ = ['SolrException', 'Solr', 'SolrConnection',
+           'Response', 'SearchHandler']
 
 _python_version = sys.version_info[0]+(sys.version_info[1]/10.0)
 
@@ -260,7 +264,24 @@ _python_version = sys.version_info[0]+(sys.version_info[1]/10.0)
 # Exceptions
 # ===================================================================
 class SolrException(Exception):
-    """ An exception thrown by solr connections """
+    """An exception thrown by solr connections.
+
+    Detailed information is provided in attributes of the exception object.
+    """
+
+    httpcode = 400
+    """HTTP response code from Solr."""
+
+    reason = None
+    """Error message from the HTTP response sent by Solr."""
+
+    body = None
+    """Response body returned by Solr.
+
+    This can contain much more information about the error, including
+    tracebacks from the Java runtime.
+    """
+
     def __init__(self, httpcode, reason=None, body=None):
         self.httpcode = httpcode
         self.reason = reason
@@ -274,28 +295,62 @@ class SolrException(Exception):
         return 'HTTP code=%s, reason=%s' % (self.httpcode, self.reason)
 
 
-# ===================================================================
-# Connection Object
-# ===================================================================
-class SolrConnection:
-    """
-    Represents a Solr connection.
+# Decorator (used below)
 
-    Designed to work with the 2.2 response format (SOLR 1.2+).
-    (though 2.1 will likely work.)
-    """
+def committing(function=None):
+
+    def wrapper(self, *args, **kw):
+        commit = kw.pop("commit", False)
+        optimize = kw.pop("optimize", False)
+        query = {}
+        if commit or optimize:
+            if optimize:
+                query["optimize"] = "true"
+            elif commit:
+                query["commit"] = "true"
+            wait_searcher = kw.pop("wait_searcher", True)
+            wait_flush = kw.pop("wait_flush", True)
+            if not wait_searcher:
+                query["waitSearcher"] = "false"
+            if not wait_flush:
+                query["waitFlush"] = "false"
+                query["waitSearcher"] = "false"
+        elif "wait_flush" in kw:
+            raise TypeError(
+                "wait_flush cannot be specified without commit or optimize")
+        elif "wait_searcher" in kw:
+            raise TypeError(
+                "wait_searcher cannot be specified without commit or optimize")
+        content = function(self, *args, **kw)
+        if content:
+            return self._update(content, query)
+        # If there's nothing to do (no content), should we issue a
+        # commit/optimize if those are requested by the options?
+
+    wrapper.__doc__ = function.__doc__
+    wrapper.__name__ = function.__name__
+    return wrapper
+
+
+# ===================================================================
+# Connection Objects
+# ===================================================================
+
+class Solr:
 
     def __init__(self, url,
                  persistent=True,
                  timeout=None,
                  ssl_key=None,
                  ssl_cert=None,
+                 http_user=None,
+                 http_pass=None,
                  post_headers={},
                  max_retries=3,
                  debug=False):
 
         """
-            url -- URI pointing to the SOLR instance. Examples:
+            url -- URI pointing to the Solr instance. Examples:
 
                 http://localhost:8080/solr
                 https://solr-server/solr
@@ -311,7 +366,10 @@ class SolrConnection:
 
             ssl_key, ssl_cert -- If using client-side key files for
                 SSL authentication,  these should be, respectively,
-                your PEM key file and certificate file
+                your PEM key file and certificate file.
+
+            http_user, http_pass -- If given, include HTTP Basic authentication 
+                in all request headers.
 
         """
 
@@ -326,14 +384,14 @@ class SolrConnection:
         self.ssl_key = ssl_key
         self.ssl_cert = ssl_cert
         self.max_retries = int(max_retries)
-        
+
         assert self.max_retries >= 0
-        
+
         kwargs = {}
-        
+
         if self.timeout and _python_version >= 2.6 and _python_version < 3:
             kwargs['timeout'] = self.timeout
-        
+
         if self.scheme == 'https':
             self.conn = httplib.HTTPSConnection(self.host,
                    key_file=ssl_key, cert_file=ssl_cert, **kwargs)
@@ -345,7 +403,7 @@ class SolrConnection:
 
         # Responses from Solr will always be in UTF-8
         self.decoder = codecs.getdecoder('utf-8')
-        
+
         # Set timeout, if applicable.
         if self.timeout and _python_version < 2.6:
             self.conn.connect()
@@ -360,184 +418,120 @@ class SolrConnection:
             self.xmlheaders['Connection'] = 'close'
 
         self.form_headers = {
-                'Content-Type':
-                'application/x-www-form-urlencoded; charset=utf-8'}
-
+            'Content-Type': 'application/x-www-form-urlencoded; charset=utf-8'}
+        
+        if http_user is not None and http_pass is not None:
+            http_auth = http_user + ':' + http_pass
+            http_auth = 'Basic ' + http_auth.encode('base64').strip()
+            self.auth_headers = {'Authorization': http_auth}
+        else:
+            self.auth_headers = {}
+        
         if not self.persistent:
             self.form_headers['Connection'] = 'close'
 
         self.debug = debug
+        self.select = SearchHandler(self, "/select")
 
     def close(self):
+        """Close the underlying HTTP(S) connection."""
         self.conn.close()
 
-    def query(self, q, fields=None, highlight=None,
-              score=True, sort=None, sort_order="asc", **params):
+
+    # Update interface.
+
+    @committing
+    def delete(self, id=None, ids=None, queries=None):
         """
-        q is the query string.
+        Delete documents by ids or queries.
 
-        fields is an optional list of fields to include. It can
-        be either a string in the format that SOLR expects, or
-        a python list/tuple of field names.   Defaults to
-        all fields. ("*")
+        Any or all of `id`, `ids`, or `queries` may be given; all
+        provided will be used.  If none are provided, no request will be
+        sent to Solr.
 
-        score indicates whether "score" should be included
-        in the field list.  Note that if you explicitly list
-        "score" in your fields value, then score is
-        effectively ignored.  Defaults to true.
+        `id` is a single value for the schema's unique id field.  `ids`
+        is an iterable of unique ids.
 
-        highlight indicates whether highlighting should be included.
-        highlight can either be False, indicating "No" (the default),
-        a list of fields in the same format as "fields" or True, indicating
-        to highlight any fields included in "fields". If True and no "fields"
-        are given, raise a ValueError.
-
-        sort is a list of fields to sort by. See "fields" for
-        formatting. Each sort element can have be in the form
-        "fieldname asc|desc" as specified by SOLR specs.
-
-        sort_order is the backward compatible way to add the same ordering
-        to all the sort field when it is not specified.
-
-        Optional parameters can also be passed in.  Many SOLR
-        parameters are in a dotted notation (e.g., hl.simple.post).
-        For such parameters, replace the dots with underscores when
-        calling this method. (e.g., hl_simple_post='</pre'>)
-
-        Returns a Response instance.
+        `queries` is an iterable of standard-syntax queries.
+        Supports commit-control arguments.
         """
+        return self._delete(id=id, ids=ids, queries=queries)
 
-        # Clean up optional parameters to match SOLR spec.
-        params = dict([(key.replace('_','.'), value)
-                      for key, value in params.items()])
-
-        if highlight:
-            params['hl'] = 'true'
-            if not isinstance(highlight, (bool, int, float)):
-                if not isinstance(highlight, basestring):
-                    highlight = ",".join(highlight)
-                params['hl.fl'] = highlight
-            else:
-                if not fields:
-                    raise ValueError("highlight is True and no fields were given")
-                elif isinstance(fields, basestring):
-                    params['hl.fl'] = [fields]
-                else:
-                    params['hl.fl'] = ",".join(fields)
-
-        if q is not None:
-            params['q'] = q
-
-        if fields:
-            if not isinstance(fields, basestring):
-                fields = ",".join(fields)
-        if not fields:
-            fields = '*'
-
-        if sort:
-            if not sort_order or sort_order not in ("asc", "desc"):
-                raise ValueError("sort_order must be 'asc' or 'desc'")
-            if isinstance(sort, basestring):
-                sort = [ f.strip() for f in sort.split(",") ]
-            sorting = []
-            for e in sort:
-                if not (e.endswith("asc") or e.endswith("desc")):
-                    sorting.append("%s %s" % (e, sort_order))
-                else:
-                    sorting.append(e)
-            sort = ",".join(sorting)
-            params['sort'] = sort
-
-        if score and not 'score' in fields.replace(',',' ').split():
-            fields += ',score'
-
-        params['fl'] = fields
-        params['version'] = self.response_version
-        params['wt'] = 'standard'
-
-        request = urllib.urlencode(params, doseq=True)
-
-        if self.debug:
-            logging.info("solrpy request: %s" % request)
-
-        try:
-            rsp = self._post(self.path + '/select',
-                              request, self.form_headers)
-            # If we pass in rsp directly, instead of using rsp.read())
-            # and creating a StringIO, then Persistence breaks with
-            # an internal python error.
-            xml = StringIO(rsp.read())
-            if self.debug:
-                logging.info("solrpy got response: %s" % xml.getvalue())
-                
-            data = parse_query_response(xml,  params=params, connection=self)
-
-        finally:
-            if not self.persistent:
-                self.close()
-
-        return data
-
-    def delete(self, id):
-        """
-        Delete a specific document by id.
-        """
-        xstr = u'<delete><id>%s</id></delete>' % escape(unicode(id))
-        return self._update(xstr)
-
+    @committing
     def delete_many(self, ids):
         """
-        Delete documents using a list of IDs.
-        """
-        [self.delete(id) for id in ids]
+        Delete documents using an iterable of ids.
 
+        This is equivalent to ``delete(ids=[ids])``.
+        Supports commit-control arguments.
+        """
+        return self._delete(ids=ids)
+
+    @committing
     def delete_query(self, query):
         """
-        Delete all documents returned by a query.
-        """
-        xstr = u'<delete><query>%s</query></delete>' % escape(query)
-        return self._update(xstr)
+        Delete all documents identified by a query.
 
-    def add(self, _commit=False, **fields):
+        This is equivalent to ``delete(queries=[query])``.
+        Supports commit-control arguments.
         """
-        Add a document to the SOLR server.  Document fields
+        return self._delete(queries=[query])
+
+    @committing
+    def add(self, doc):
+        """
+        Add a document to the Solr server.  Document fields
         should be specified as arguments to this function
 
-        Example:
-            connection.add(id="mydoc", author="Me")
+        Example::
+
+            doc = {"id": "mydoc", "author": "Me"}
+            connection.add(doc)
+
+        Supports commit-control arguments.
         """
         lst = [u'<add>']
-        self.__add(lst, fields)
+        self.__add(lst, doc)
         lst.append(u'</add>')
-        xstr = ''.join(lst)
-        if not _commit:
-            return self._update(xstr)
-        else:
-            self._update(xstr)
-            return self.commit()
+        return ''.join(lst)
 
-    def add_many(self, docs, _commit=False):
+    @committing
+    def add_many(self, docs):
         """
-        Add several documents to the SOLR server.
+        Add several documents to the Solr server.
 
-        docs -- a list of dicts, where each dict is a document to add
-            to SOLR.
+        `docs`
+            An iterable of document dictionaries.
+
+        Supports commit-control arguments.
         """
         lst = [u'<add>']
         for doc in docs:
             self.__add(lst, doc)
         lst.append(u'</add>')
-        xstr = ''.join(lst)
-        if not _commit:
-            return self._update(xstr)
-        else:
-            self._update(xstr)
-            return self.commit()
+        return ''.join(lst)
 
     def commit(self, wait_flush=True, wait_searcher=True, _optimize=False):
         """
-        Issue a commit command to the SOLR server.
+        Issue a commit command to the Solr server.
+
+        `wait_flush` and `wait_searcher` have the same interpretations as
+        the like-name `commit-control arguments`_.
+
         """
+        return self._commit("commit", wait_flush, wait_searcher)
+
+    def optimize(self, wait_flush=True, wait_searcher=True):
+        """
+        Issue an optimize command to the Solr server.
+
+        `wait_flush` and `wait_searcher` have the same interpretations as
+        the like-name `commit-control arguments`_.
+
+        """
+        return self._commit("optimize", wait_flush, wait_searcher)
+
+    def _commit(self, verb, wait_flush, wait_searcher):
         if not wait_searcher:  #just handle deviations from the default
             if not wait_flush:
                 options = 'waitFlush="false" waitSearcher="false"'
@@ -545,58 +539,24 @@ class SolrConnection:
                 options = 'waitSearcher="false"'
         else:
             options = ''
-
-        if _optimize:
-            xstr = u'<optimize %s/>' % options
-        else:
-            xstr = u'<commit %s/>' % options
-
+        xstr = u'<%s %s/>' % (verb, options)
         return self._update(xstr)
 
-    def optimize(self, wait_flush=True, wait_searcher=True, ):
-        """
-        Issue an optimize command to the SOLR server.
-        """
-        self.commit(wait_flush, wait_searcher, _optimize=True)
-
-    def raw_query(self, **params):
-        """
-        Issue a query against a SOLR server.
-
-        Return the raw result.  No pre-processing or
-        post-processing happends to either
-        input parameters or responses
-        """
-
-        # Clean up optional parameters to match SOLR spec.
-        params = dict([(key.replace('_','.'), value)
-                       for key, value in params.items()])
-
-        request = urllib.urlencode(params, doseq=True)
+    # Helper methods.
 
+    def _update(self, request, query=None):
+        selector = '%s/update%s' % (self.path, qs_from_items(query))
         try:
-            rsp = self._post(self.path+'/select',
-                              request, self.form_headers)
-            data = rsp.read()
-        finally:
-            if not self.persistent:
-                self.close()
-
-        return data
-
-    def _update(self, request):
-
-        try:
-            rsp = self._post(self.path + '/update',
-                              request, self.xmlheaders)
+            rsp = self._post(selector, request, self.xmlheaders)
             data = rsp.read()
         finally:
             if not self.persistent:
                 self.close()
 
         # Detect old-style error response (HTTP response code
-        # of 200 with a non-zero status.
-        if data.startswith('<result status="') and not data.startswith('<result status="0"'):
+        # of 200 with a non-zero status).
+        starts = data.startswith
+        if starts('<result status="') and not starts('<result status="0"'):
             data = self.decoder(data)[0]
             parsed = parseString(data)
             status = parsed.documentElement.getAttribute('status')
@@ -634,11 +594,30 @@ class SolrConnection:
                     escape(unicode(value)))))
         lst.append('</doc>')
 
+    def _delete(self, id=None, ids=None, queries=None):
+        """
+        Delete a specific document by id.
+        """
+        if not ids:
+            ids = []
+        if id:
+            ids.insert(0, id)
+        lst = []
+        for id in ids:
+            lst.append(u'<id>%s</id>\n' % escape(unicode(id)))
+        for query in (queries or ()):
+            lst.append(u'<query>%s</query>\n' % escape(unicode(query)))
+        if lst:
+            lst.insert(0, u'<delete>\n')
+            lst.append(u'</delete>')
+            return ''.join(lst)
+
     def __repr__(self):
-        return ('<SolrConnection (url=%s, '
-                'persistent=%s, post_headers=%s, reconnects=%s)>') % (
-            self.url, self.persistent,
-            self.xmlheaders, self.reconnects)
+        return (
+            '<%s (url=%s, persistent=%s, post_headers=%s, reconnects=%s)>'
+            % (self.__class__.__name__,
+               self.url, self.persistent,
+               self.xmlheaders, self.reconnects))
 
     def _reconnect(self):
         self.reconnects += 1
@@ -651,23 +630,200 @@ class SolrConnection:
                 self.conn.sock.sock.settimeout(self.timeout)
 
     def _post(self, url, body, headers):
+        _headers = self.auth_headers.copy()
+        _headers.update(headers)
         attempts = self.max_retries + 1
         while attempts > 0:
             try:
-                self.conn.request('POST', url, body.encode('UTF-8'), headers)
+                self.conn.request('POST', url, body.encode('UTF-8'), _headers)
                 return check_response_status(self.conn.getresponse())
             except (socket.error,
                     httplib.ImproperConnectionState,
                     httplib.BadStatusLine):
                     # We include BadStatusLine as they are spurious
                     # and may randomly happen on an otherwise fine
-                    # SOLR connection (though not often)
+                    # Solr connection (though not often)
                 self._reconnect()
                 attempts -= 1
                 if attempts <= 0:
                     raise
 
 
+class SolrConnection(Solr):
+    """
+    Represents a Solr connection.
+
+    Designed to work with the 2.2 response format (Solr 1.2+),
+    though will likely work with 2.1 responses as well.
+    """
+
+    # Backward compatible update interfaces.
+
+    def add(self, _commit=False, **fields):
+        """
+        Add or update a single document with field values given by
+        keyword arguments.
+
+        The `_commit` argument is treated specially, causing an immediate
+        commit if present.  It may be specified either positionally or as
+        a keyword.  If `_commit` is true, the commit will be issued as
+        part of the same HTTP request to the Solr server.
+
+        Example::
+
+            connection.add(id="mydoc", author="Me")
+
+        This is equialent to ``solr.Solr.add(fields, commit=_commit)``.
+        """
+        return Solr.add_many(self, [fields], commit=_commit)
+
+    def add_many(self, docs, _commit=False):
+        """
+        Add or update multiple documents. with field values for each given
+        by dictionaries in the sequence `docs`.
+
+        The `_commit` argument is treated specially, causing an immediate
+        commit if present.  It may be specified either positionally or as
+        a keyword.  If `_commit` is true, the commit will be issued as
+        part of the same HTTP request to the Solr server.
+
+        Example::
+
+            doc1 = {...}
+            doc2 = {...}
+            connection.add_many([doc1, doc2], _commit=True)
+
+        This is equialent to ``solr.Solr.add_many(docs, commit=_commit)``.
+        """
+        return Solr.add_many(self, docs, commit=_commit)
+
+    # Backward compatible query interfaces.
+
+    def query(self, *args, **params):
+        return self.select(*args, **params)
+
+    def raw_query(self, **params):
+        return self.select.raw(**params)
+
+
+class SearchHandler(object):
+
+    def __init__(self, conn, relpath="/select"):
+        self.conn = conn
+        self.selector = conn.path + relpath
+
+    def __call__(self, q=None, fields=None, highlight=None,
+                 score=True, sort=None, sort_order="asc", **params):
+        """
+        q is the query string.
+
+        fields is an optional list of fields to include. It can
+        be either a string in the format that SOLR expects, or
+        a python list/tuple of field names.   Defaults to
+        all fields. ("*")
+
+        score indicates whether "score" should be included
+        in the field list.  Note that if you explicitly list
+        "score" in your fields value, then score is
+        effectively ignored.  Defaults to true.
+
+        highlight indicates whether highlighting should be included.
+        highlight can either be False, indicating "No" (the default),
+        a list of fields in the same format as "fields" or True, indicating
+        to highlight any fields included in "fields". If True and no "fields"
+        are given, raise a ValueError.
+
+        sort is a list of fields to sort by. See "fields" for
+        formatting. Each sort element can have be in the form
+        "fieldname asc|desc" as specified by SOLR specs.
+
+        sort_order is the backward compatible way to add the same ordering
+        to all the sort field when it is not specified.
+
+        Optional parameters can also be passed in.  Many SOLR
+        parameters are in a dotted notation (e.g., hl.simple.post).
+        For such parameters, replace the dots with underscores when
+        calling this method. (e.g., hl_simple_post='</pre'>)
+
+        Returns a Response instance.
+        """
+        # Optional parameters with '_' instead of '.' will be converted
+        # later by raw_query().
+
+        if highlight:
+            params['hl'] = 'true'
+            if not isinstance(highlight, (bool, int, float)):
+                if not isinstance(highlight, basestring):
+                    highlight = ",".join(highlight)
+                params['hl_fl'] = highlight
+            else:
+                if not fields:
+                    raise ValueError("highlight is True and no fields were given")
+                elif isinstance(fields, basestring):
+                    params['hl_fl'] = [fields]
+                else:
+                    params['hl_fl'] = ",".join(fields)
+
+        if q is not None:
+            params['q'] = q
+
+        if fields:
+            if not isinstance(fields, basestring):
+                fields = ",".join(fields)
+        if not fields:
+            fields = '*'
+
+        if sort:
+            if not sort_order or sort_order not in ("asc", "desc"):
+                raise ValueError("sort_order must be 'asc' or 'desc'")
+            if isinstance(sort, basestring):
+                sort = [ f.strip() for f in sort.split(",") ]
+            sorting = []
+            for e in sort:
+                if not (e.endswith("asc") or e.endswith("desc")):
+                    sorting.append("%s %s" % (e, sort_order))
+                else:
+                    sorting.append(e)
+            sort = ",".join(sorting)
+            params['sort'] = sort
+
+        if score and not 'score' in fields.replace(',',' ').split():
+            fields += ',score'
+
+        params['fl'] = fields
+        params['version'] = self.conn.response_version
+        params['wt'] = 'standard'
+
+        xml = self.raw(**params)
+        return parse_query_response(StringIO(xml),  params, self)
+
+    def raw(self, **params):
+        """
+        Issue a query against a SOLR server.
+
+        Return the raw result.  No pre-processing or post-processing
+        happens to either input parameters or responses.
+        """
+        # Clean up optional parameters to match SOLR spec.
+        params = dict([(key.replace('_','.'), value)
+                       for key, value in params.items()])
+        request = urllib.urlencode(params, doseq=True)
+        conn = self.conn
+        if conn.debug:
+            logging.info("solrpy request: %s" % request)
+
+        try:
+            rsp = conn._post(self.selector, request, conn.form_headers)
+            data = rsp.read()
+            if conn.debug:
+                logging.info("solrpy got response: %s" % data)
+        finally:
+            if not conn.persistent:
+                conn.close()
+
+        return data
+
+
 # ===================================================================
 # Response objects
 # ===================================================================
@@ -682,14 +838,14 @@ class Response(object):
           results -- a list of matching documents. Each list item will
               be a dict.
     """
-    def __init__(self, connection):
+    def __init__(self, query):
         # These are set in ResponseContentHandler.endElement()
         self.header = {}
         self.results = []
 
         # These are set by parse_query_response().
         # Used only if .next_batch()/previous_batch() is called
-        self._connection = connection
+        self._query = query
         self._params = {}
 
     def _set_numFound(self, value):
@@ -741,7 +897,7 @@ class Response(object):
         """
         Load the next set of matches.
 
-        By default, SOLR returns 10 at a time.
+        By default, Solr returns 10 at a time.
         """
         try:
             start = int(self.results.start)
@@ -753,7 +909,7 @@ class Response(object):
         params['start'] = start
         q = params['q']
         del params['q']
-        return self._connection.query(q, **params)
+        return self._query(q, **params)
 
     def previous_batch(self):
         """
@@ -774,13 +930,13 @@ class Response(object):
         params['rows'] = rows
         q = params['q']
         del params['q']
-        return self._connection.query(q, **params)
+        return self._query(q, **params)
 
 
 # ===================================================================
 # XML Parsing support
 # ===================================================================
-def parse_query_response(data, params, connection):
+def parse_query_response(data, params, query):
     """
     Parse the XML results of a /select call.
     """
@@ -791,7 +947,7 @@ def parse_query_response(data, params, connection):
     if handler.stack[0].children:
         response = handler.stack[0].children[0].final
         response._params = params
-        response._connection = connection
+        response._query = query
         return response
     else:
         return None
@@ -936,9 +1092,9 @@ def check_response_status(response):
 
 
 # -------------------------------------------------------------------
-# Datetime extensions to parse/generate SOLR date formats
+# Datetime extensions to parse/generate Solr date formats
 # -------------------------------------------------------------------
-# A UTC class, for parsing SOLR's returned dates.
+# A UTC class, for parsing Solr's returned dates.
 class UTC(datetime.tzinfo):
     """
     UTC timezone.
@@ -957,7 +1113,7 @@ utc = UTC()
 
 def utc_to_string(value):
     """
-    Convert datetimes to the subset of ISO 8601 that SOLR expects.
+    Convert datetimes to the subset of ISO 8601 that Solr expects.
     """
     value = value.astimezone(utc).isoformat()
     if '+' in value:
@@ -969,7 +1125,7 @@ def utc_from_string(value):
     """
     Parse a string representing an ISO 8601 date.
     Note: this doesn't process the entire ISO 8601 standard,
-    onle the specific format SOLR promises to generate.
+    onle the specific format Solr promises to generate.
     """
     try:
         if not value.endswith('Z') and value[10] == 'T':
@@ -984,4 +1140,19 @@ def utc_from_string(value):
         return datetime.datetime(year, month, day, hour,
             minute, second, microsecond, utc)
     except ValueError:
-        raise ValueError ("'%s' is not a valid ISO 8601 SOLR date" % value)
+        raise ValueError ("'%s' is not a valid ISO 8601 Solr date" % value)
+
+def qs_from_items(query):
+    # This deals with lists of values since multiple filter queries can
+    # be used for a single request.
+    qs = ''
+    if query:
+        sep = '?'
+        for k, v in query.items():
+            k = urllib.quote(k)
+            if isinstance(v, basestring):
+                v = [v]
+            for s in v:
+                qs += "%s%s=%s" % (sep, k, urllib.quote_plus(s))
+                sep = '&'
+    return qs
